package be.iminds.ilabt.jfed.gts.model.impl;

import be.iminds.ilabt.jfed.gts.model.GtsFactory;

/**
 * User: twalcari
 * Date: 6/5/2015
 * Time: 16:22
 */
public class FXGtsFactory implements GtsFactory {

    public FXGtsHost createGtsHost() {
        return new FXGtsHost();
    }

    public FXGtsOfx createGtsOfx() {
        return new FXGtsOfx();
    }

    public FXGtsExternalNetwork createGtsExternalNetwork() {
        return new FXGtsExternalNetwork();
    }

    public FXGtsLink createGtsLink() {
        return new FXGtsLink();
    }

    public FXGtsPort createGtsPort() {
        return new FXGtsPort();
    }

    public FXGtsType createGtsType() {
        return new FXGtsType();
    }
    public FXGtsType createGtsType(String name) {
        return new FXGtsType(name);
    }


}
